import random
import pytest
from ex1 import concat, concat2, reversed, drop, get_item, hd, head, last, nth, reversed2, take, zip, unzip
from strategies import arbitrary_int, arbitrary_list, arbitrary_lowercase_letter

# random.seed(1)

# list of identifiers for the generated test cases (default number is 10)
test_ids = ["test_" + str(i) for i in range(10)]


@pytest.mark.parametrize(
    ("l"),
    [arbitrary_list(
        lambda: arbitrary_int(0, 10),
        arbitrary_lowercase_letter)
     for _ in test_ids],
    ids=test_ids,
)
def test_hd_property3(l: list[int]) -> None:
    match l:
        case [h, *_]:
            result = hd(l)
            assert result == h
        case _:
            with pytest.raises(ValueError):
                hd(l)


def test_get_item() -> None:
    expected: tuple[int, list[int]] = (10, [21, 32, 43])
    actual = get_item([10, 21, 32, 43])
    assert actual is not None
    assert expected == actual


def test_last() -> None:
    expected = 43
    actual = last([10, 21, 32, 43])
    assert expected == actual


def test_nth() -> None:
    expected = 32
    actual = nth([10, 21, 32, 43], 2)
    assert expected == actual


def test_take() -> None:
    expected = [10, 21]
    actual = take([10, 21, 32, 43], 2)
    assert expected == actual


def test_drop() -> None:
    expected = [32, 43]
    actual = drop([10, 21, 32, 43], 2)
    assert expected == actual


def test_reversed() -> None:
    expected = [43, 32, 21, 10]
    actual = reversed([10, 21, 32, 43])
    assert expected == actual


@pytest.mark.parametrize(
    ("l"),
    [arbitrary_list(
        lambda: arbitrary_int(0, 10),
        lambda: arbitrary_int(-10, 10))
     for _ in test_ids],
    ids=test_ids,
)
def test_reversed_property(l: list[int]) -> None:
    # print("\n", "l =", l)
    match l:
        case [h, *t]:
            result = reversed(l)
            assert result == reversed(t) + [h]
        case _:
            result = reversed(l)
            assert result == []


@pytest.mark.parametrize(
    ("l"),
    [arbitrary_list(
        lambda: arbitrary_int(0, 10),
        lambda: arbitrary_int(-10, 10))
     for _ in test_ids],
    ids=test_ids,
)
def test_reversed2_property(l: list[int]) -> None:
    match l:
        case [h, *t]:
            result = reversed2(l)
            assert result == reversed2(t) + [h]
        case _:
            result = reversed2(l)
            assert result == []


@pytest.mark.parametrize(("l"), [
    ([]),
    ([1, 2, 3]),
    ([1, 2, 3, 4, 5]),
])
def test_reversed2_property2(l: list[int]) -> None:
    result = reversed2(l)
    assert result == reversed(l)


def test_concat() -> None:
    expected = [1, 2, 3, 4, 5, 6]
    actual = concat([[1, 2], [], [3, 4, 5], [6]])
    assert expected == actual


@pytest.mark.parametrize(
    ("l"),
    [arbitrary_list(
        lambda: arbitrary_int(0, 15),
        lambda: arbitrary_list(
            lambda: arbitrary_int(0, 5),
            lambda: arbitrary_int(-10, 10)))
        for _ in test_ids],
    ids=test_ids,
)
def test_concat_property(l: list[list[int]]) -> None:
    print("\n", "l =", l)
    result = concat2(l)
    assert result == concat(l)


@pytest.mark.parametrize(
    ("l1", "l2"),
    [(arbitrary_list(
        lambda: 10,
        lambda: arbitrary_int(-10, 10)),
      arbitrary_list(
        lambda: 10,
        lambda: arbitrary_lowercase_letter()))
        for _ in test_ids],
    ids=test_ids,
)
def test_zip_property(l1: list[int], l2: list[str]) -> None:
    print("\n", "l1 =", l1, "l2 =", l2)
    l = zip(l1, l2)
    (r1, r2) = unzip(l)
    assert r1 == l1 and r2 == l2
