from collections.abc import Callable, Iterator
import random


type Supplier[T] = Callable[[], T]


def arbitrary_int(start: int, end: int) -> int:
    """ Returns an arbitrary integer between 'start' and 'end' (both inclusive).
        Raises ValueError if 'start' is greater than 'end' """
    return random.randint(start, end)


def arbitrary_int_10_20() -> int:
    return arbitrary_int(10, 20)

# lambda: arbitrary_int(10, 20)


def arbitrary_char(start: str, end: str) -> str:
    """ Returns an arbitrary character between 'start' and 'end' (both inclusive).
        Raises ValueError if 'start' is greater than 'end' and TypeError if 
        both arguments are not chars. """
    i = arbitrary_int(ord(start), ord(end))
    return chr(i)


def arbitrary_bool() -> bool:
    """ Returns either True or False. """
    i = arbitrary_int(0, 1)
    return bool(i)


def arbitrary_char_from(possible_chars: str) -> str:
    """ Returns an arbitrary character that occur in the given string.
        Raises ValueError if the given string is empty """
    i = arbitrary_int(0, len(possible_chars) - 1)
    return possible_chars[i]


def arbitrary_uppercase_letter() -> str:
    """ Returns an arbitrary upper case ASCII letter ('A'-'Z'). """
    return arbitrary_char('A', 'Z')


def arbitrary_lowercase_letter() -> str:
    """ Returns an arbitrary lower case ASCII letter ('a'-'z'). """
    return arbitrary_char('a', 'z')


def arbitrary_digit() -> str:
    """ Returns an arbitrary decimal digit ('0'-'9'). """
    return arbitrary_char('0', '9')


def arbitrary_str(arbitrary_size: Supplier[int], arbitrary_char: Supplier[str]) -> str:
    """ Returns a string of arbitrary size containing arbitrary chars. """
    size = arbitrary_size()
    s: str = ""
    while (len(s) < size):
        s += arbitrary_char()
    return s


# arbitrary_str(arbitrary_int_10_20 , arbitrary_uppercase_letter)
# arbitrary_str(lambda: arbitrary_int(10, 20) , arbitrary_uppercase_letter)

def arbitrary_from[T](values: list[T]) -> T:
    """ Returns an arbitrary value taken from the given list.
        Raises ValueError if the given list is empty. """
    i = arbitrary_int(0, len(values) - 1)
    return values[i]


def arbitrary_list[T](arbitrary_size: Supplier[int], arbitrary_item: Supplier[T]) -> list[T]:
    """ Returns a list of arbitrary size containing arbitrary items. """
    size = arbitrary_size()
    l: list[T] = []
    while (len(l) < size):
        l += [arbitrary_item()]
    return l


def arbitrary_dict[K, V](arbitrary_size: Supplier[int], arbitrary_key: Supplier[K],
                         arbitrary_value: Supplier[V]) -> dict[K, V]:
    """ Returns a dict of arbitrary size mapping arbitrary keys to arbitrary values. """
    ...
    size = arbitrary_size()
    d: dict[K, V] = {}
    while (len(d) < size):
        d |= {arbitrary_key(): arbitrary_value()}
    return d


def main() -> None:
    # print(arbitrary_uppercase_letter())
    # print(arbitrary_lowercase_letter())
    # print(arbitrary_str(arbitrary_int_10_20, arbitrary_lowercase_letter))
    # print(arbitrary_list(arbitrary_int_10_20, arbitrary_lowercase_letter))
    print(arbitrary_dict(lambda: 5,
                         lambda: arbitrary_int(1, 100),
                         lambda: arbitrary_str(arbitrary_int_10_20, arbitrary_lowercase_letter)))


if __name__ == "__main__":
    main()
